#include <slang.h>
#include <termio.h>

#include <signal.h>
#include <algorithm>

#include "gfx.hh"
#include "gfx_term.hh"
#include "bstringfun.hh"
#include "error.hh"
#include "time.hh"

static TerminalScreen* term = 0;

extern "C" { 
    static void SlangUpdate(int)
    {
        std::pair<int,int> oldsize = term->GetWinSize();
        
        SLtt_get_screen_size();
        signal(SIGWINCH, SlangUpdate);
        
        term->SizeChanged(oldsize);
    }
    static void SigIntHandler(int)
    {
        Term_Done();
        signal(SIGINT, SIG_DFL);
        raise(SIGINT);
    }
    /*static void SigStopHandler(int)
    {
        SLsmg_suspend_smg();
        signal(SIGTSTP, SigStopHandler);
        raise(SIGSTOP);
    }*/
    static void SigContHandler(int)
    {
        //if(!SLsmg_resume_smg()) return;
        Term_Done();
        Term_Init();
        if(term)
        {
            SLsmg_touch_screen();
            term->ReRenderVpage();
        }
    }
}

static struct termio back;

void Term_Init()
{
    ioctl(0, TCGETA, &back);
    SLtt_get_terminfo();
    SLkp_init();
    SLang_init_tty(-1/*don't change intr*/, 1/*enable xon/xoff*/, 0);
    signal(SIGINT, SigIntHandler);
    //signal(SIGTSTP, SigStopHandler);
    signal(SIGCONT, SigContHandler);
    SLtty_set_suspend_state(1);
}

void Term_Done()
{
    SLang_reset_tty();
    ioctl(0, TCSETA, &back);
}

TerminalScreen::TerminalScreen()
    : Screen(),
      pages(8),
      apage(0), vpage(0),
      fgc(7), bgc(0)
{
    SLsmg_init_smg ();
    SLsig_unblock_signals();
    term = this;
    signal(SIGWINCH, SlangUpdate);

    for(unsigned b=0; b<16; ++b)
        for(unsigned f=0; f<16; ++f)
        {
            unsigned bswapped = (b & 0xA) + 4*(b&1) + (b&4)/4;
            unsigned fswapped = (f & 0xA) + 4*(f&1) + (f&4)/4;
            
            SLtt_set_color_fgbg(b*16+f, fswapped, bswapped);
        }

    SLsmg_Newline_Behavior = SLSMG_NEWLINE_MOVES;
    
    SetPages(8);
    ResetViewPrint();
}

TerminalScreen::~TerminalScreen()
{
    /*std::pair<int,int> pos = GetWinSize();
    pos.first  = 1;
    SetCursorPos(pos);*/
    
    signal(SIGWINCH, SIG_DFL);
    term = 0;

    //SLsmg_reset_smg();
    SLang_reset_tty();
}

std::pair<int,int> TerminalScreen::GetWinSize()
{
    return std::pair<int,int> (SLtt_Screen_Cols, SLtt_Screen_Rows);
}

void TerminalScreen::SetPages(int npages)
{
    std::pair<int,int> siz = GetWinSize();
    
    const unsigned short clear_with = 0x20 | ((bgc*16 + fgc) << 8);
    pages.resize(npages);

    for(size_t n=0; n<npages; ++n)
        pages[n].contents.resize(siz.first * siz.second, clear_with);
    
    if(apage >= npages) SetAPage(0);
    if(vpage >= npages) SetVPage(0);
}

void TerminalScreen::SetAPage(int pageno)
{
    apage = pageno;
}

void TerminalScreen::SetVPage(int pageno)
{
    if(vpage != pageno)
    {
        vpage = pageno;
        
        ReRenderVpage();
    }
}

std::pair<int,int> TerminalScreen::GetCursorPos()
{
    std::pair<int,int> result = pages[apage].cursorpos;
    result.first  += 1;
    result.second += 1;
    return result;
}

void TerminalScreen::SetCursorPos(const std::pair<int,int>& newpos)
{
    const std::pair<int,int> size = GetWinSize();
    const std::pair<int,int>& win  = GetViewPrint();
    if(newpos.first < 1
    || newpos.second < 1
    || newpos.first > size.first
    || newpos.second > size.second
    || newpos.second < win.first
    || newpos.second > win.second) B_ERROR(5); // illegal function call
    
    std::pair<int,int> actualpos = newpos;
    actualpos.first  -= 1;
    actualpos.second -= 1;
    
    pages[apage].cursorpos = actualpos;
}

std::pair<int,int> TerminalScreen::ReadScreen(int x,int y)
{
    const std::pair<int,int> size = GetWinSize();
    if(x < 1
    || y < 1
    || x > size.first
    || y > size.second) B_ERROR(5); // illegal function call
    
    unsigned short res = pages[apage].contents[(x-1) + (y-1) * size.first];
    return std::pair<int,int>( res & 0xFF, res >> 8 );
}

void TerminalScreen::SetWinSize(const std::pair<int,int>& newsize, bool approximate)
{
}

char TranslateChar(char c, bool& swapattr)
{
    switch( (unsigned char) c)
    {
        case 219: swapattr=true; return 0x20;
        case 220: return 'z';
        case 223: return '';
    }
    return c;
}

void TerminalScreen::OutS(const char* buffer, size_t nbytes)
{
    int width = GetWinSize().first;
    std::pair<int,int> win  = GetViewPrint();
    win.first  -= 1;
    win.second -= 1;
    
    Page& p = pages[apage];
    std::pair<int,int>& cursor = p.cursorpos;
    
    for(size_t b=0; b<nbytes; ++b)
    {
        bool swapattr = false;
        char c = TranslateChar(buffer[b], swapattr);
        
        if(c == '\r' || c == '\n') // both do a cr+lf in QB
        {
        CRLF:;
            cursor.first = 0;
            cursor.second += 1;
            if(cursor.second > win.second)
            {
                ScrollRegion(win.first, win.second, -1);
                cursor.second -= 1;
            }
        }
        else if(c == '\b')
        {
            if(cursor.first > 0) --cursor.first;
        }
        else if(c == '\t')
        {
            while(cursor.first & 7) ++cursor.first;
        WrapCheck:
            if(cursor.first >= width) goto CRLF;
        }
        else
        {
            if(swapattr)
                p.contents[cursor.first + width*cursor.second] =
                    ((bgc + (fgc*16)) << 8) | (unsigned char)c;
            else
                p.contents[cursor.first + width*cursor.second] =
                    ((fgc + (bgc*16)) << 8) | (unsigned char)c;
            ++cursor.first;
            goto WrapCheck;
        }
    }
}

void TerminalScreen::SetColor(int fg, int bg, int border)
{
    if(fg > 15 || bg > 15) B_ERROR(5); // illegal function call
    if(fg >= 0) fgc = fg;
    if(bg >= 0) bgc = bg;
    SLsmg_set_color(fgc + (bgc*16));
}

void TerminalScreen::ClearEntireScreen()
{
    ScrollRegion(0, GetWinSize().second-1, GetWinSize().second);
    
    int firstline = GetViewPrint().first;
    
    pages[apage].cursorpos.first  = 0;
    pages[apage].cursorpos.second = 0;

    if(apage == vpage) ReRenderVpage();
}

void TerminalScreen::ClearText(int firstline, int lastline)
{
    int scroll_amount = (lastline-firstline)+2;
    ScrollRegion(firstline-1, lastline-1, -scroll_amount);
    
    pages[apage].cursorpos.first  = 0;
    pages[apage].cursorpos.second = firstline-1;
    
    if(apage == vpage) ReRenderVpage();
}

void TerminalScreen::SizeChanged(const std::pair<int,int>& oldsize)
{
    const std::pair<int,int> newsize = GetWinSize();

    const unsigned short clear_with = 0x20 | ((bgc*16+fgc) << 8);
    
    for(size_t n=0, npages=pages.size(); n<npages; ++n)
    {
        Page& p = pages[n];
        
        std::vector<unsigned short> newcontents(newsize.first * newsize.second);
        
        for(int pos=0, x=0; x<newsize.first; ++x)
            for(int y=0; y<newsize.second; ++y)
            {
                newcontents[pos++] =
                    (  x < oldsize.first
                    && y < oldsize.second)
                    ? p.contents[x + oldsize.first * y]
                    : clear_with;
            }
        newcontents.swap(p.contents);
        if(p.cursorpos.first  >= newsize.first)  p.cursorpos.first  = newsize.first-1;
        if(p.cursorpos.second >= newsize.second) p.cursorpos.second = newsize.second-1;
    }
    ReRenderVpage();
}

void TerminalScreen::ReRenderVpage()
{
    const Page& p = pages[vpage];

    const std::pair<int,int> size = GetWinSize();

    // refresh screen contents
    for(int y=0; y<size.second; ++y)
    {
        int oldcolor=-1;
        int xbegin=0;
        std::string str;
        for(int x=0; x<size.first; ++x)
        {
        ReX:;
            unsigned short c = p.contents[x + size.first * y];
            int color = c >> 8;
            
            int eol_len = 0;
            for(int x2=x; x2<size.first; ++x2, ++eol_len)
                if(p.contents[x2 + size.first * y] != (0x20 | (color << 8)))
                    break;
            
            bool do_eol_optimization = x + eol_len >= size.first || eol_len >= 20;

            if(color != oldcolor || do_eol_optimization)
            {
                if(!str.empty())
                {
                    SLsmg_gotorc(y, xbegin);
                    SLsmg_set_color(oldcolor);
                    SLsmg_write_nchars(const_cast<char*> (str.c_str()), str.size());
                    str.clear();
                }
                xbegin = x;
                oldcolor = color;
            }
            
            if(do_eol_optimization)
            {
                SLsmg_gotorc(y, x);
                SLsmg_set_color(color);
                SLsmg_erase_eol();
                
                x += eol_len;
                xbegin = x;
                if(x >= size.first) break;
                goto ReX;
            }
        
        NotEOLpossible:
            str += (char)(c & 0xFF);
        }
        if(!str.empty())
        {
            SLsmg_gotorc(y, xbegin);
            SLsmg_set_color(oldcolor);
            SLsmg_write_nchars(const_cast<char*> (str.c_str()), str.size());
        }
    }
    
    SLsmg_set_color(fgc + (bgc*16));
    SLsmg_gotorc(p.cursorpos.second,
                 p.cursorpos.first);
    SLsmg_refresh();
}

void TerminalScreen::ScrollRegion(int firstline, int lastline, int direction)
{
    Page& p = pages[apage];

    int width = GetWinSize().first;
    const unsigned short clear_with = ((bgc*16 + fgc) << 8) | 0x20;
    std::vector<unsigned short> empty_line(width, clear_with);
    
    if(direction < 0) /* scroll up */
    {
        for(int y=firstline; y<=lastline; ++y)
        {
            int src_line = y - direction;
            if(src_line < firstline || src_line > lastline)
                std::copy(&empty_line[0],
                          &empty_line[width],
                          &p.contents[y * width]);
            else
                std::copy(&p.contents[src_line * width],
                          &p.contents[(src_line+1)*width],
                          &p.contents[y * width]);
        }
    }
    else if(direction > 0) /* scroll down */
    {
        for(int y=lastline; y>=firstline; --y)
        {
            int src_line = y - direction;
            if(src_line < firstline || src_line > lastline)
                std::copy(&empty_line[0],
                          &empty_line[width],
                          &p.contents[y * width]);
            else
                std::copy(&p.contents[src_line * width],
                          &p.contents[(src_line+1)*width],
                          &p.contents[y * width]);
        }
    }
}

void TerminalScreen::ScreenPOKE(int x,int y, int colorflag, int value)
{
    if(x < 1 || y < 1) return;
    std::pair<int,int> siz = GetWinSize();
    if(x > siz.first || y > siz.second) return;
    
    Page& p = pages[0];
    
    unsigned short& cell = p.contents[ (y-1) * siz.first + (x-1) ];
    
    if(colorflag)
        cell = (cell & 0xFF) + (value << 8);
    else
        cell = (cell & 0xFF00) + value;
    
    /*if(vpage == 0) ReRenderVpage();
    */
}

const BAS_STR B_FuncINKEY_S(bool wait)
{
    static double last_timer = 0;
    double time = B_FuncTIMER();
    int wait_length = 0; // deciseconds
    /*if(time == last_timer)
        wait_length = 1;
    else
        last_timer = time;*/
    
    if(!wait && !SLang_input_pending(wait_length))
    {
        return "";
    }
    unsigned int key = SLang_getkey();
    switch(key)
    {
        case '':
        case SL_KEY_UP: { static const char Buf[2] = {0,'H'}; return std::string(Buf,Buf+2); }
        case '':
        case SL_KEY_DOWN: { static const char Buf[2] = {0,'P'}; return std::string(Buf,Buf+2); }
        case '':
        case SL_KEY_LEFT: { static const char Buf[2] = {0,'K'}; return std::string(Buf,Buf+2); }
        case '':
        case SL_KEY_RIGHT: { static const char Buf[2] = {0,'M'}; return std::string(Buf,Buf+2); }
        case '':
        case SL_KEY_PPAGE: { static const char Buf[2] = {0,'I'}; return std::string(Buf,Buf+2); }
        case '':
        case SL_KEY_NPAGE: { static const char Buf[2] = {0,'Q'}; return std::string(Buf,Buf+2); }
        case '':
        case SL_KEY_HOME: { static const char Buf[2] = {0,'G'}; return std::string(Buf,Buf+2); }
        case '':
        case SL_KEY_END: { static const char Buf[2] = {0,'O'}; return std::string(Buf,Buf+2); }
        case '':
        case SL_KEY_BACKSPACE: { static const char Buf[1] = {8}; return std::string(Buf,Buf+1); }
        case '\r': case '\n':
        case SL_KEY_ENTER: { static const char Buf[1] = {13}; return std::string(Buf,Buf+1); }
        case '':
        case SL_KEY_DELETE: { static const char Buf[1] = {127}; return std::string(Buf,Buf+1); }
        
        /*
            R: insert?
            S: delete
            Q: pagedown
            G: home
            O: end
            I: pageup
        */
        default: if(key >= 256) key = '?';
    }
    return B_FuncCHR_S(key);
}

void B_StmtSLEEP(double duration)
{
    B_RerenderTextScreen();
    
    /* Waits until either the duration has elapsed
     * or an ON event occurs.
     */
    while(duration > 1e-6 || duration == -1.0)
    {
        int waitlen = duration * 10;
        if(duration == -1.0) waitlen = 100;
        if(SLang_input_pending(waitlen)) return;
        
        if(duration != -1.0)
            duration -= waitlen / 10.0;
        
        usleep((int)(duration * 1e6));
        duration -= (int)(duration * 1e6) / 1e6;
    }
}

void B_RerenderTextScreen()
{
    if(term) term->ReRenderVpage();
}
