#include "stream.hh"
#include "error.hh"
#include "bstringfun.hh"

#include <math.h>  // for fabs and log10. log10 wouldn't be in std::.
#include <cctype>
#include <stdio.h> // for snprintf. Probably wouldn't be in std:: anyway.

/*
                     Characters that format a numeric expression    
 #    Digit position.                    | -     Placed after digit, prints
 .    Decimal point position.            |       trailing sign for negative    
 ,    Placed left of the decimal point,  |       numbers.    
      prints a comma every third digit.  | $$    Prints leading $.    
 +    Position of number sign.           | **    Fills leading spaces with *.
^^^^  Prints in exponential format.      | **$   Combines ** and $$.

                     Characters used to format a string expression    
 &    Prints entire string.              | \ \   Prints first n characters,
 !    Prints only the first character    |       where n is the number of    
      of the string.                     |       blanks between slashes + 2.

                     Characters used to print literal characters    
 _    Prints the following formatting    |       Any character not in this    
      character as a literal.            |       table is printed as a
                                         |       literal.
NOTES:
    $$ allocates two digit positions just like ## does
              except in scientific notation
    -  terminates the digit sequence (also counts as a character of sequence length)
       Note: If number is not negative, counts as a space
    .  also allocates one digit position
    ,  so does this. Note: Affects only once; position irrelvant.
    +  also allocates one digit position:
        Note: If comes ahead the sequence, indicates that
              sign is to be placed on left side of sequence
              If comes in middle of sequence, terminates the
              digit sequence (still allocates one slot).
    
    Example on using ^^^^:   (Note: D for doubles, E for others)
    
     |+#.##^^^^  for 123456 prints:
     |+1.23D+05
     |
     |+##.##^^^^  for 123456 prints:
     |+12.35D+04
     |
     |###.##^^^^^  for 123456 prints:
     | 12.35D+004
     |
     |###.##^^^^^+  for 123456 prints:
     |123.46D+003+
     |
     |$$#.##^^^^   for 123456 prints:
     | $1.23D+05
     |
     |**#.##^^^^   for 123456 prints:
     |*12.34D+04
     |
     |**#.##^^^^  for -123456 prints:
     |-12.34D+04
     |
     |**$.##^^^^   for 123456 prints:
     |*$1.23D+05
     |
     |**$.##^^^^  for -123456 prints:
     |-$1.23D+05
     |
     |#.##^^^^   for 123456 prints:
     |0.12D+06
     |
     |#.##^^^^  for -123456 prints:
     |-.12D+06

*/

UsingFormatter::UsingFormatter(const BAS_STR& f): format(), pos(0)
{
    CreateFormat(f);
}
UsingFormatter::~UsingFormatter()
{
}

void UsingFormatter::DoParam(PrintStreamBase& out, const BAS_LNG v)
{
    // 64-bit IEEE double has enough range to cover a 32-bit signed integer,
    // so to save effort and code, we handle it as a double.
    DoParam(out, (BAS_DBL)v, 'E');
}

void UsingFormatter::DoParam(PrintStreamBase& out, const BAS_DBL v)
{
    DoParam(out, v, 'D');
}

void UsingFormatter::DoParam(PrintStreamBase& out, const BAS_SGN v)
{
    DoParam(out, (BAS_DBL)v, 'E');
}

void UsingFormatter::OutputNonscientific(
    PrintStreamBase& out,
    const BAS_DBL v,
    const UsingFormatter::FormatItem& format)
{
    std::vector<char> Buffer;

    int bytes_required = snprintf(&Buffer[0], 0,
        "%.*f",
            (int) format.num_trailing_chars,
            fabs( v ));
    bytes_required += 8;

    Buffer.resize(bytes_required);
    
    int begin = 4;

    snprintf(&Buffer[begin], bytes_required - begin,
        "%.*f",
            (int) format.num_trailing_chars,
            fabs( v ));

    if(format.leading_dollar) Buffer[--begin] = '$';
    if(format.sign_type == FormatItem::NormalSignLeft && v < 0) Buffer[--begin] = '-';
    if(format.sign_type == FormatItem::AlwaysSignLeft) Buffer[--begin] = (v < 0 ? '-' : '+');
    
    std::string num_basis = &Buffer[begin];
    if(format.sign_type == FormatItem::NegativeSignRight) num_basis += (v < 0 ? '-' : ' ');
    if(format.sign_type == FormatItem::AlwaysSignRight) num_basis += (v < 0 ? '-' : '+');
    
    if(format.comma_every_third_digit)
    {
        std::string::size_type decimpoint = num_basis.find('.');
        if(decimpoint == num_basis.npos) decimpoint = num_basis.size();
        
        while(decimpoint > 3)
        {
            if(std::isdigit(num_basis[decimpoint-1])
            && std::isdigit(num_basis[decimpoint-2])
            && std::isdigit(num_basis[decimpoint-3])
            && std::isdigit(num_basis[decimpoint-4])
              )
            {
                num_basis.insert(decimpoint-3, 1, ',');
                decimpoint -= 3;
                continue;
            }
            break;
        }
    }
    
    std::string::size_type decimpoint = num_basis.find('.');
    if(decimpoint == num_basis.npos) decimpoint = num_basis.size();
    
    if(decimpoint == format.num_leading_chars+1
    && num_basis[0]=='-'
    && num_basis[1]=='0'
    && num_basis[2]=='.')
    {
        --decimpoint;
        num_basis.erase(1,1);
    }
    
    if(decimpoint > format.num_leading_chars)
    {
        num_basis.insert(0, 1, '%'); // indicate overflow
    }
    else
    {
        num_basis.insert(0, format.num_leading_chars-decimpoint,
                            format.fill_stars ? '*' : ' ');
    }
    
    out << num_basis;
}

void UsingFormatter::DoParam(PrintStreamBase& out, const BAS_DBL v, char ExpChar)
{
    FlushLiterals(out, true);
    
    const FormatItem& format = GetFormat();
    if(format.type != FormatItem::Numeric) B_ERROR(13); // type mismatch
    
    if(format.scientific)
    {
        //FIXME
        // Note: QB ignores comma in scientific mode.
        
        int leading_decimals = format.num_leading_chars;
        if(format.sign_type == FormatItem::AlwaysSignLeft
        || format.sign_type == FormatItem::NormalSignLeft)
        {
            --leading_decimals;
        }
        if(format.leading_dollar)
        {
            --leading_decimals;
        }
        
        // L = 10's exponent to divide v with,
        // so that result is 0.0 <= v < 1.0
        int L = (int) ((1 + 1e-8) + log10( v==0.0 ? 1.0 : fabs(v) ));
        
        L -= leading_decimals;
        
        double value = v / pow(10.0, L);
        
        FormatItem tmpformat = format;
        tmpformat.scientific = false;
        if(format.sign_type == FormatItem::NegativeSignRight
        || format.sign_type == FormatItem::AlwaysSignRight)
        {
            tmpformat.sign_type = FormatItem::NormalSignLeft;
            value = fabs(value);
        }
        
        OutputNonscientific(out, value, tmpformat);
        
        char expo[512];
        sprintf(expo, "%c%+0*d", ExpChar, 1 + format.scientific, L);
        
        out << expo;

        if(format.sign_type == FormatItem::NegativeSignRight) out << (v < 0 ? '-' : ' ');
        if(format.sign_type == FormatItem::AlwaysSignRight) out << (v < 0 ? '-' : '+');
    }
    else
    {
        OutputNonscientific(out, v, format);
    }

    FlushLiterals(out, false);
}

void UsingFormatter::DoParam(PrintStreamBase& out, const BAS_STR& v)
{
    FlushLiterals(out, true);
    
    const FormatItem& format = GetFormat();
    if(format.type != FormatItem::String) B_ERROR(13); // type mismatch
    
    if(format.length == -1)
        out << v;
    else
    {
        size_t n_eat    = format.length;
        size_t n_spaces = 0;
        
        if(n_eat > v.size())
        {
            n_spaces = n_eat - v.size();
            n_eat    = v.size();
        }
        
        if(n_eat    > 0) out << v.substr(0, n_eat);
        if(n_spaces > 0) out << B_FuncSTRING_S(n_spaces, 32);
    }
    
    FlushLiterals(out, false);
}

void UsingFormatter::CreateFormat(const BAS_STR& f)
{
    if(false)
    {
    IllCall:
        format.clear(); // this will do IllCall at runtime
        return;
    }
    
    std::string literal;
    const char* const s = f.c_str();
    size_t spos = 0, slen = f.size();
    
    #define FlushLiteral() do { \
        if(!literal.empty()) \
        { \
            FormatItem tmp; \
            tmp.type    = FormatItem::Literal; \
            tmp.literal = literal; \
            format.push_back(tmp); \
            literal.clear(); \
        } \
    } while(0)
    
    while(spos < slen)
    {
        #define remain (slen-spos)
        #define c      (s[spos])
        
        if(c == '_' && remain >= 2)
        {
            literal += s[spos+1];
            spos += 2; continue;
        }
        if(c == '&')
        {
            FlushLiteral();
            FormatItem tmp;
            tmp.type   = FormatItem::String;
            tmp.length = -1;
            format.push_back(tmp);
            spos += 1; continue;
        }
        if(c == '!')
        {
            FlushLiteral();
            FormatItem tmp;
            tmp.type   = FormatItem::String;
            tmp.length = 1;
            format.push_back(tmp);
            spos += 1; continue;
        }
        if(c == '\\' && remain >= 2)
        {
            int strlen = 2;
            for(size_t p=spos+1; p<slen; ++p)
            {
                if(s[p] == '\\') break;
                if(s[p] != ' ') goto not_valid_backslash_thing;
                ++strlen;
            }
            
            { FlushLiteral();
            FormatItem tmp;
            tmp.type   = FormatItem::String;
            tmp.length = strlen;
            format.push_back(tmp);
            spos += strlen; continue;
            }
        not_valid_backslash_thing:;
        }
        
        if(1) // numeric processing
        {
            size_t saved_spos = spos;
            
            #define RequireNextC(n) \
                do { spos += n; if(spos >= slen) goto IllCall; \
                   } while(0)
            #define AcquireNextC(n) \
                do { spos += n; if(spos >= slen) goto DoneNumFmt; \
                   } while(0)
        
            // be optimistic, start creating the numeral structure.
            FormatItem tmp;
            tmp.numdata = 0; // fill all numeric data by zero.
            
            if(c == '+') // heading +
            {
                tmp.sign_type         = FormatItem::AlwaysSignLeft;
                tmp.num_leading_chars += 1;
                RequireNextC(1);
            }
            if(c == '$' && remain >= 2 && s[spos+1]=='$')
            {
                tmp.leading_dollar    = true;
                tmp.num_leading_chars += 2;
                AcquireNextC(2);
            }
            else if(c == '*' && remain >= 2 && s[spos+1]=='*')
            {
                int n = 2;
                if(remain >= 3 && s[spos+2]=='$')
                {
                    tmp.leading_dollar    = true;
                    n = 3;
                }
                tmp.fill_stars        = true;
                tmp.num_leading_chars += n;
                AcquireNextC(n);
            }
            for(;;)
            {
                if(c == ',' && !tmp.decimal_point)
                {
                    tmp.comma_every_third_digit = true;
                    tmp.num_leading_chars += 1;
                    AcquireNextC(1);
                }
                else if(c == '.' && !tmp.decimal_point && !tmp.scientific)
                {
                    tmp.decimal_point = true;
                    AcquireNextC(1);
                }
                else if(c == '-' && tmp.sign_type != FormatItem::AlwaysSignLeft)
                {
                    tmp.sign_type = FormatItem::NegativeSignRight;
                    AcquireNextC(1);
                    break; // end of the sequence.
                }
                else if(c == '+' && tmp.sign_type != FormatItem::AlwaysSignLeft)
                {
                    tmp.sign_type = FormatItem::AlwaysSignRight;
                    AcquireNextC(1);
                    break; // end of the sequence.
                }
                else if(c == '#' && !tmp.scientific)
                {
                    if(tmp.decimal_point)
                        ++tmp.num_trailing_chars;
                    else
                        ++tmp.num_leading_chars;
                    AcquireNextC(1);
                }
                else if(remain >= 4 && c         == '^'
                                    && s[spos+1] == '^'
                                    && s[spos+2] == '^'
                                    && s[spos+3] == '^')
                {
                    int n = 4;
                    if(remain >= 5 && s[spos+4] == '^') n=5;
                    tmp.scientific = n-2;
                    AcquireNextC(n);
                }
                else
                    break;
            }
        DoneNumFmt:;
            //fprintf(stderr, "DoneNumFmt(%s)\n", s+spos);
            if(tmp.numdata != 0) // if something was filled
            {
                if(tmp.num_leading_chars == 0
                && tmp.num_trailing_chars == 0)
                {
                    // illegal pattern.
                    spos = saved_spos;
                }
                else
                {
                    FlushLiteral();
                    tmp.type = FormatItem::Numeric;
                    format.push_back(tmp);
                    continue; // accept
                }
            }
            #undef AcquireNextC
            #undef RequireNextC
        }
        
        // anything else:
        literal += c;
        spos += 1;
        
        #undef remain
        #undef c
    }
    FlushLiteral();
  #if 0
    fprintf(stderr, "%u format items\n",(unsigned)format.size());
    for(size_t n=0; n<format.size(); ++n)
    {
        FormatItem& fmt = format[n];
        switch(fmt.type)
        {
            case FormatItem::Literal:
                fprintf(stderr, "[%d]: Literal \"%s\"\r\n",
                    (int)n, fmt.literal.c_str());
                break;
            case FormatItem::String:
                fprintf(stderr, "[%d]: String %d\r\n",
                    (int)n, fmt.length);
                break;
            case FormatItem::Numeric:
                fprintf(stderr, "[%d]: Numeric %d.%d sci=%d %s %s %s %s %d\r\n",
                    (int)n,
                    fmt.num_leading_chars,
                    fmt.num_trailing_chars,
                    fmt.scientific,
                    fmt.fill_stars ? "stars" : "no stars",
                    fmt.decimal_point ? "point" : "no point",
                    fmt.comma_every_third_digit ? "comma" : "no comma",
                    fmt.leading_dollar ? "dollar" : "no dollar",
                    (int)fmt.sign_type
                );
                break;
           default:
               fprintf(stderr, "[%d]: eh?\r\n", (int)n);
        }
    }
    fflush(stderr);
  #endif
}

const UsingFormatter::FormatItem& UsingFormatter::GetFormat()
{
    if(format.empty())
    {
    Error:
        B_ERROR(5); // illegal function call
    }
    
    const FormatItem& result = format[pos];
    if(++pos == format.size()) pos = 0;
    return result;
}

void UsingFormatter::FlushLiterals(PrintStreamBase& out, bool allow_wrap)
{
    if(format.empty())
    {
    Error:
        B_ERROR(5); // illegal function call
    }
    
    if(!allow_wrap && pos == 0) return;
    
    size_t origpos = pos;
    while(format[pos].type == FormatItem::Literal)
    {
        out << format[pos].literal;
        if(++pos == format.size()) pos = 0;
        if(pos == origpos) goto Error; // full loop without finding a parameter
        
        if(pos == 0 && !allow_wrap) break;
    }
}
