#include <cstdio>
#include <cstring>

#include "stream.hh"
#include "error.hh"
#include "input.hh"
#include "gfx.hh" 

#include <map>

/* File writing */
class QbFPstream: public PrintStreamBase
{
    FILE* fp;
    int xpos;
public:
    QbFPstream(int fd) : fp(fdopen(fd, "wb")), xpos(0) { }
    
    virtual unsigned GetXpos() const
    {
        return xpos;
    }
    virtual unsigned GetXsize() const
    {
        return 65535;
    }
    virtual void DoCRLF()
    {
        fputc('\n', fp);
    }
    virtual void Write(const void* buf, size_t n)
    {
        size_t res = fwrite(buf, 1, n, fp);
        // FIXME: error checking
        fflush(fp);
    }
};

/* File input */
class QbFIstream: public InputImplement
{
    FILE* fp;
public:
    QbFIstream(int fd) : fp(fdopen(fd, "rb")) { }

    virtual void ReadLine(BAS_STR& target)
    {
        target = "";
        for(;;)
        {
            int c = InputChar();
            if(c == '\r') continue;
            if(c == '\n' || c == -1) break;
            target += (char) c;
        }
    }

    virtual void BeginInput()
    {
        InputError = false;
        GotEOL     = false;
        // File:
        //   Do nothing. InputTo reads file.
    }

    virtual bool EndInputOkp()
    {
        // File:
        //   Return false if InputTo failed, true otherwise.
        if(InputError) return false;
        return !FinishInput();
    }

    virtual BAS_STR Input_N(size_t n)
    {
        BAS_STR buf(n, '\0');
        size_t res = fread(&buf[0], 1, n, fp);
        if(res == 0)
        {
            if(ferror(fp)) { clearerr(fp); B_ERROR_Errno(); }
        }
        buf.erase(buf.begin() + res, buf.end());
        return buf;
    }

    virtual int InputChar()
    {
     retry:;
        int res = fgetc(fp);
        if(res == '\r') goto retry;
        if(res == EOF || res == '\n') { GotEOL = true; return -1; }
        return res;
    }

    virtual void GotOverflow()
    {
        B_ERROR(6); // overflow
        InputError = true;
    }
};

/* Console I/O */
class QbConsoleStream: public PrintStreamBase, public InputImplement
{
    BAS_STR InputLineBuffer;
    unsigned InputLineBufferPos;
public:
    QbConsoleStream() : PrintStreamBase(), InputImplement() { }
    
    virtual unsigned GetXpos() const
    {
        return B_FuncPOS();
    }
    virtual unsigned GetXsize() const
    {
        return B_FuncWIDTH();
    }
    virtual void DoCRLF()
    {
        GFX_OutS("\r", 1);
    }
    virtual void Write(const void* buf, size_t n)
    {
        GFX_OutS( (const char*) buf, n);
    }

    virtual void ReadLine(BAS_STR& target)
    {
        target = "";
        int strpos  = 0;
        int curspos = B_FuncPOS();
        bool insertmode = true;
        for(;;)
        {
            BAS_STR inkey = B_FuncINKEY_S(true);
            if(inkey == "\n" || inkey == "\r") break;
            if(inkey[0] == '\0')
                switch(inkey[1])
                {
                    case 'K': // left
                        if(strpos > 0) { --strpos; --curspos; goto relocate; }
                        break;
                    case 'M': // right
                        if(strpos < target.size())
                                       { ++strpos; ++curspos; goto relocate; }
                        break;
                    case 'O': // end
                        curspos += target.size() - strpos;
                        strpos = target.size(); goto relocate;
                    case 'Q': // home
                        curspos -= strpos; strpos = 0; goto relocate;
                    default: continue;
                }
            else if(inkey.size() == 1)
                switch(inkey[0])
                {
                    case 8: // backspace
                        if(strpos > 0)
                            { target.erase(--strpos, 1); --curspos;
                              B_StmtLOCATE(-1, curspos);
                              GFX_OutS(&target[strpos], target.size()-strpos);
                              GFX_OutS(" ", 1);
                              goto relocate;
                            }
                        continue;
                }

            target.insert(strpos, inkey);

            GFX_OutS(&target[strpos], target.size()-strpos);

            if(insertmode)
            {
                strpos += inkey.size();
                curspos += inkey.size();
            }
            else
                target.erase(strpos + inkey.size(), inkey.size());
        relocate:
            B_StmtLOCATE(-1, curspos);
        }
        
        if(strpos != target.size())
        {
            curspos += target.size() - strpos;
            strpos = target.size();
            B_StmtLOCATE(-1, curspos);
        }
        // note: does not go to next line.
    }

    virtual void BeginInput()
    {
        InputError = false;
        GotEOL     = false;
        // Console:
        //   Read line into buffer. InputTo queries buffer.
        ReadLine(InputLineBuffer); // note: keeps cursor on same line
        InputLineBufferPos=0;
    }

    virtual bool EndInputOkp()
    {
        // Console:
        //   Return false if InputTo failed.
        //   Return false if buffer is not empty.
        //   Return true otherwise.

        if(InputError) return false;
        return !FinishInput();
    }

    virtual BAS_STR Input_N(size_t n)
    {
        // No echoing
        // FIXME: Don't take control chars.
        BAS_STR res;
        for(; n > 0; --n) res += B_FuncINKEY_S(true);
        return res;
    }

    virtual int InputChar()
    {
        if(InputLineBufferPos < InputLineBuffer.size())
        {
            return InputLineBuffer[InputLineBufferPos++];
        }
        GotEOL = true;
        return -1;
    }

    virtual void GotOverflow()
    {
        InputError = true;
        *this << "Overflow\n";
    }
};

struct FileDescriptor
{
    QbFPstream fpout;
    QbFIstream fpin;
};
static std::map<int/*qbasic filenum*/, FileDescriptor> fdmap;

PrintStreamBase& GET_STREAM_OUTPUT(int filenumber)
{
    std::map<int, FileDescriptor>::iterator
        i = fdmap.find(filenumber);

    if(i == fdmap.end())
        B_ERROR(52); // bad file number
    
    return i->second.fpout;
}

InputStreamBase& GET_STREAM_INPUT(int filenumber)
{
    std::map<int, FileDescriptor>::iterator
        i = fdmap.find(filenumber);

    if(i == fdmap.end())
        B_ERROR(52); // bad file number
    
    return i->second.fpin;
}

static QbConsoleStream cons;

PrintStreamBase& GET_STDOUT()
{
    static bool tty_out = isatty(1);
    if(!tty_out) { static QbFPstream out(1); return out; }
    return cons;
}
InputStreamBase& GET_STDIN()
{
    static bool tty_in = isatty(0);
    if(!tty_in) { static QbFIstream in(0); return in; }
    return cons;
}
